/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) Robert Kaye 2003
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: tagger.cpp,v 1.22 2003/06/27 23:08:25 robert Exp $

----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <deque>
#include "tunepimp.h"
#include "../config.h"

#ifdef HAVE_LIBREADLINE
#  include <readline/readline.h>
#  include <readline/history.h>
#endif

const char *fileStatusText[] =
{
    "Unrecognized",
    "Recognized",
    "Pending",
    "TRM Lookup",
    "TRM Collision",
    "File Lookup",
    "User Selection",
    "Verified",
    "Deleted",
    "Error"
};


#ifdef HAVE_LIBREADLINE
                                      
/*
 * Read a line from standard input using libreadline and put it into
 * the history buffer.
 */
char *getCommand(void)                
{                                     
	static char *line = NULL;

        if ( line != NULL )           
                free(line);           
                                      
        line = readline("tag> ");     

	/* line can be NULL if the user typed Ctrl-D. */
	if ( line == NULL )
		line = strdup("q");	/* quit */
	else
		if ( line[0] != '\0' )
			add_history(line);
                                      
        return line;                  
}                                     
                                      
#else                                 

static char cmdLine[256];
char *getCommand(void)
{
    printf("tag> ");
    fflush(stdout);
    if (fgets(cmdLine, 256, stdin))
        cmdLine[strlen(cmdLine) - 1] = 0;

    return cmdLine;
}

#endif /* HAVE_LIBREADLINE */

void printList(TunePimp *pimp)
{
    vector<int>           ids;
    vector<int>::iterator i;
    string                fileName;
    Track                *track;

    pimp->getFileIds(ids);
    if (ids.size() == 0)
        printf("No files in the tagger.\n");
    else
        for(i = ids.begin(); i != ids.end(); i++)
        {
            track = pimp->getTrack(*i);
            if (!track)
            {
                printf("Can't get track %d info\n", *i);
                continue;
            }

            track->lock();
            track->getFileName(fileName);
            printf("%d: %s, %3d%%: %s\n", *i, 
                                   fileStatusText[(int)track->getStatus()],
                                   track->getSimilarity(),
                                   fileName.c_str());
            if (track->getStatus() == eError)
            {
                string err;

                track->getError(err);
                printf("  Error: %s\n", err.c_str());
            }
            track->unlock();
            pimp->releaseTrack(track);
        }

    printf("\n%d unsubmitted TRMs.\n\n", pimp->getNumUnsubmitted());
}

void printTrack(TunePimp &pimp, int id)
{
    Track                *track;
    string                fileName, trm;
    Metadata              data;

    track = pimp.getTrack(id);
    if (!track)
    {
        printf("Can't get track %d\n\n", id);
        return;
    }

    track->lock();
    track->getFileName(fileName);
    printf("%d: %s: %s\n", id, 
                           fileStatusText[(int)track->getStatus()],
                           fileName.c_str());

    track->getTRM(trm);
    printf("       TRM: %s\n", trm.c_str());
    printf("Similarity: %d\n", track->getSimilarity());
    printf("   Changed: %d\n\n", track->hasChanged());

    track->getLocalMetadata(data);
    printf("File metadata:\n");
    printf("   Artist: %s\n", data.artist.c_str());
    printf("    Album: %s\n", data.album.c_str());
    printf("    Track: %s\n", data.track.c_str());
    printf(" TrackNum: %d\n", data.trackNum);
    printf(" Duration: %ld\n", data.duration);
    printf(" ArtistId: %s\n", data.artistId.c_str());
    printf("  AlbumId: %s\n", data.albumId.c_str());
    printf("  TrackId: %s\n", data.trackId.c_str());
    printf("       VA: %d\n\n", data.variousArtist);

    track->getServerMetadata(data);
    printf("Server metadata:\n");
    printf("   Artist: %s\n", data.artist.c_str());
    printf("    Album: %s\n", data.album.c_str());
    printf("    Track: %s\n", data.track.c_str());
    printf(" TrackNum: %d\n", data.trackNum);
    printf(" Duration: %ld\n", data.duration);
    printf(" ArtistId: %s\n", data.artistId.c_str());
    printf("  AlbumId: %s\n", data.albumId.c_str());
    printf("  TrackId: %s\n", data.trackId.c_str());
    printf("       VA: %d\n", data.variousArtist);
    if (data.albumType != eAlbumType_Error)
    {
        string type;

        convertFromAlbumType(data.albumType, type);
        printf("AlbumType: %s\n", type.c_str());
    }

    printf("\n");
    track->unlock();

    pimp.releaseTrack(track);
}

void printArtistResult(Track *track, vector<TPResult *> &results)
{
    vector<TPResult *>::iterator  i;
    TPArtistResult               *result;
    string                        fileName;
    int                           count;

    track->lock();
    track->getFileName(fileName);
    track->unlock();

    printf("Artists for %s:\n--------------------------------------------\n", fileName.c_str());
    for(i = results.begin(), count = 0; i != results.end(); i++, count++)
    {
         result = (TPArtistResult *)*i;
         printf("%2d. %d%% %-32s %-32s\n", count,
               result->getRelevance(),
               result->getName().c_str(), 
               result->getSortName().c_str());
    }
    printf("\n");
}

void printAlbumResult(Track *track, vector<TPResult *> &results)
{
    vector<TPResult *>::iterator  i;
    TPAlbumResult                *result;
    TPArtistResult                artist;
    string                        fileName;
    int                           count;

    track->lock();
    track->getFileName(fileName);
    track->unlock();

    for(i = results.begin(), count = 0; i != results.end(); i++, count++)
    {
        result = (TPAlbumResult *)*i;
        if (i == results.begin())
        {
            artist = result->getArtist();
            printf("Artist: %s (%s)\n", artist.getName().c_str(), artist.getSortName().c_str());
            printf("Albums for: %s\n---------------------------------------------\n", fileName.c_str());
        }

        printf("%2d. %d%% %-40s (%d trks %d cd)\n", count,
            result->getRelevance(),
            result->getName().c_str(), 
            result->getNumTracks(),
            result->getNumCDIndexIds());
    } 
    printf("\n");
}

void printTrackResult(Track *track, vector<TPResult *> &results)
{
    vector<TPResult *>::iterator  i;
    TPAlbumTrackResult           *result;
    TPArtistResult                artist;
    TPAlbumResult                 album;
    string                        fileName;
    int                           count;

    track->lock();
    track->getFileName(fileName);
    track->unlock();

    for(i = results.begin(), count = 0; i != results.end(); i++, count++)
    {
        result = (TPAlbumTrackResult *)*i;
        if (i == results.begin())
        {
            artist = result->getArtist();
            printf("Artist: %s (%s)\n", artist.getName().c_str(), artist.getSortName().c_str());
            printf("Album/Tracks for: %s\n---------------------------------------------\n", fileName.c_str());
        }
        album = result->getAlbum();
        printf("%2d %d%% %-32s %-32s (",
                   count,
                   result->getRelevance(),
                   result->getName().c_str(),
                   album.getName().c_str());
        if (album.getNumTracks() >= 0)
            printf("%d trks ", album.getNumTracks());
        if (album.getNumCDIndexIds() >= 0)
            printf("%d cd ", album.getNumCDIndexIds());

        printf("%d trm)\n", result->getNumTRMIds());
    }
    printf("\n");
}

void printResults(TunePimp &pimp, int id)
{
    Track                       *track;
    string                       fileName, trm;
    Metadata                     data;
    LookupFile::Result           type;
    vector<TPResult *>           results;

    track = pimp.getTrack(id);
    if (!track)
    {
        printf("Can't get track %d\n\n", id);
        return;
    }

    track->lock();
    track->getFileName(fileName);

    track->getResults(type, results);
    if (results.size() == 0)
    {
        printf("No results available for this track.\n\n");
    }
    else
    {
        switch(type)
        {
            case LookupFile::eArtistList:
                printArtistResult(track, results);
                break;
            case LookupFile::eAlbumList:
                printAlbumResult(track, results);
                break;
            case LookupFile::eTrackList:
                printTrackResult(track, results);
                break;
            default:
                printf("No clue what this result is.\n");
        }
    }

    track->unlock();

    pimp.releaseTrack(track);
}

void toggleVA(TunePimp &pimp, int id)
{
    Track                *track;
    string                fileName, trm;
    Metadata              data;

    track = pimp.getTrack(id);
    if (!track)
    {
        printf("Can't get track %d\n\n", id);
        return;
    }

    track->lock();
    if (track->getStatus() != eUserSelection && track->getStatus() != eRecognized)
    {
        track->unlock();
        pimp.releaseTrack(track);
        printf("Only tracks waiting for a user selection or recognized tracks "
               "can be looked up as various artist.\n");
        return;
    }
    else
    {
        track->getLocalMetadata(data);
        data.artistId = "89ad4ac3-39f7-470e-963a-56509c546377";
        track->setLocalMetadata(data);
        track->setStatus(eFileLookup);
    }
    track->unlock();

    pimp.wake(track);
    pimp.releaseTrack(track);
}

void setNewMetadata(TunePimp &pimp, int id)
{
    Track                *track;
    string                fileName, trm;
    Metadata              data;
    char                  temp[256];

    track = pimp.getTrack(id);
    if (!track)
    {
        printf("Can't get track %d\n", id);
        return;
    }

    track->lock();
    track->getServerMetadata(data);
    track->unlock();

    printf("Artist: ");
    fflush(stdout);
    fgets(temp, 255, stdin);
    temp[strlen(temp) - 1] = 0;
    data.artist = string(temp);
    
    printf("Album: ");
    fflush(stdout);
    fgets(temp, 255, stdin);
    temp[strlen(temp) - 1] = 0;
    data.album = string(temp);
    
    printf("Track: ");
    fflush(stdout);
    fgets(temp, 255, stdin);
    temp[strlen(temp) - 1] = 0;
    data.track = string(temp);

    data.trackNum = 0;
    data.duration = 0;
    data.artistId = "";
    data.albumId = "";
    data.trackId = "";

    track->lock();
    track->setServerMetadata(data);
    track->setStatus(eFileLookup);
    track->unlock();

    pimp.wake(track);
    pimp.releaseTrack(track);
}

void selectTrack(TunePimp &pimp, int id, int result)
{
    Track                *track;
    string                fileName, trm;
    Metadata              data;

    track = pimp.getTrack(id);
    if (!track)
    {
        printf("Can't get track %d\n", id);
        return;
    }
    pimp.selectResult(track, result);
    pimp.releaseTrack(track);
}

class Callback : public TPCallback
{
    public:

                 Callback(TunePimp *pimp) { this->pimp = pimp; };
        virtual ~Callback(void) {};

        void notify(TPCallbackEnum type, int fileId)
        {
            string  fileName, notice;
            Track  *track;

            track = pimp->getTrack(fileId);
            if (track)
            {
                track->getFileName(fileName);
                pimp->releaseTrack(track);
            }

            switch(type)
            {
                case tpFileAdded:
                    notice = string("File added: ") + fileName.c_str();
                    break;
                case tpFileRemoved:
                    notice = string("File removed: ") + fileName.c_str();
                    break;
                case tpFileChanged:
                    notice = string(fileStatusText[(int)track->getStatus()]) +
                             string(": ") + fileName;

                    break;
                default:
                    notice = string("Unknown notify event.");
            }
            notices.push_back(notice);
        };
        bool getNotice(string &notice)
        {
            if (notices.size() == 0)
                return false;

            notice = notices.front();
            notices.pop_front();

            return true;
        };
        void status(const string &status) 
        {
            //printf("Status: %s\n", status.c_str());
        };

    private:

        TunePimp      *pimp;
        deque<string>  notices;
};

int main(int argc, char *argv[])
{
    TunePimp        pimp("tp_tagger", VERSION);
    Callback        callback(&pimp);
    vector<string>  extList;
    vector<string>::iterator j;
    char           *cmd;
    string          notice, outputDir("music");
    bool            debugOpt = false;
    bool            renameOpt = true;
    bool            moveOpt = true;
    bool            clearTagsOpt = false;
    string          server("test.musicbrainz.org"), fileMask, variousFileMask;
    int             collThreshold = 80;
    int             autoSaveThreshold = 95;

    fileMask = "%sortname/%album/%sortname-%album-%0num-%track";
    variousFileMask = "Various Artists/%album/%album-%0num-%sortname-%track";

    pimp.setUseUTF8(0);
    pimp.setCallback(&callback);
    pimp.setDestDir(outputDir);
    pimp.setRenameFiles(true);
    pimp.setMoveFiles(true);
    pimp.setTopSrcDir(".");
    pimp.setClearTags(clearTagsOpt);
    pimp.setFileMask(fileMask);
    pimp.setVariousFileMask(variousFileMask);
    pimp.setAutoSaveThreshold(autoSaveThreshold);
    pimp.setTRMCollisionThreshold(collThreshold);

    printf("Supported file extensions\n");
    pimp.getSupportedExtensions(extList);
    for(j = extList.begin(); j != extList.end(); j++)
        printf("%s ", (*j).c_str());
    printf("\nEnter ? or h for help\n\n");

    for(int i = 1; i < argc; i++)
       pimp.add(argv[i]);

    pimp.setServer(server.c_str(), 80);
    for(;;)
    {
        while(callback.getNotice(notice))
            printf("%s\n", notice.c_str());

        cmd = getCommand();
        if (*cmd == 0)
            continue;

        if (*cmd == 'q')
        {
            printf("cleaning up...\n");
            return 0;
        }
        if (*cmd == 'l')
        {
            printList(&pimp);
            continue;
        }
        if (*cmd == 'a')
        {
            pimp.add(string(cmd + 2));
            continue;
        }
        if (*cmd == 'd')
        {
            pimp.addDir(string(cmd + 2));
            continue;
        }
        if (*cmd == 'n')
        {
            if (strlen(cmd) <= 2)
                printf("Usage: n <file #>\n");
            else
                setNewMetadata(pimp, atoi(cmd + 2));
            continue;
        }
        if (*cmd == 'm')
        {
            if (strlen(cmd) <= 2)
                printf("Usage: m <file #>\n");
            else
                pimp.misidentified(atoi(cmd + 2));
            continue;
        }
        if (*cmd == 'i')
        {
            if (strlen(cmd) <= 2)
                printf("Usage: i <file #>\n");
            else
                pimp.identifyAgain(atoi(cmd + 2));
            continue;
        }
        if (*cmd == 'p')
        {
            if (strlen(cmd) <= 2)
                printf("Usage: p <file #>\n");
            else
                printTrack(pimp, atoi(cmd + 2));
            continue;
        }
        if (*cmd == 'r')
        {
            if (strlen(cmd) <= 2)
                printf("Usage: r <file #>\n");
            else
                printResults(pimp, atoi(cmd + 2));
            continue;
        }
        if (*cmd == 'v')
        {
            if (strlen(cmd) <= 2)
                printf("Usage: v <file #>\n");
            else
                toggleVA(pimp, atoi(cmd + 2));
            continue;
        }
        if (*cmd == 'w')
        {
            bool ret;

            if (strlen(cmd) <= 2)
                ret = pimp.writeTags();
            else
            {
                vector<int> vec;

                vec.push_back(atoi(cmd + 2));
                ret = pimp.writeTags(&vec);
            }
            continue;
        }
        if (*cmd == 'c')
        {
            int fileId, result;

            if (sscanf(cmd, " %*c %d %d", &fileId, &result) == 2)
                selectTrack(pimp, fileId, result);
            else
                printf("usage: c <file #> <choice>\n");
            continue;
        }
        if (*cmd == 's')
        {
            pimp.setUserInfo("test", "mb");
            if (pimp.submitTRMs() == tpOk)
                printf("Submitted ok.\n");
            else
            {
                string err;

                pimp.getError(err);
                printf("Submit error: %s\n", err.c_str());
            }
            continue;
        }
        if (*cmd == 'o')
        {
            moveOpt = !moveOpt;
            pimp.setMoveFiles(moveOpt);
            printf("Move is: %s\n", moveOpt ? "on" : "off");
            continue;
        }
        if (*cmd == 'e')
        {
            renameOpt = !renameOpt;
            pimp.setRenameFiles(renameOpt);
            printf("Rename is: %s\n", renameOpt ? "on" : "off");
            continue;
        }
        if (*cmd == 'b')
        {
            debugOpt = !debugOpt;
            pimp.setDebug(debugOpt);
            printf("Debug: %s\n", debugOpt ? "on" : "off");
            continue;
        }
        if (strncmp(cmd, "tc", 2) == 0)
        {
            if (strlen(cmd) <= 2)
                printf("Usage: tc <trm collision threshold> (0-100)\n");
            else
            {
                collThreshold = atoi(cmd + 2);
                pimp.setTRMCollisionThreshold(collThreshold);
            }
            continue;
        }
        if (strncmp(cmd, "ta", 2) == 0)
        {
            if (strlen(cmd) <= 2)
                printf("Usage: ta <auto save threshold> (0-100)\n");
            else
            {
                autoSaveThreshold = atoi(cmd + 2);
                pimp.setAutoSaveThreshold(autoSaveThreshold);
            }
            continue;
        }
        if (*cmd == 't')
        {
            char newPath[255];

            if (sscanf(cmd, " %*c %[^\n]", newPath) == 1)
            {
                outputDir = newPath;
                pimp.setDestDir(outputDir);
                printf("New output dir is [%s]\n", newPath);
            }
            else
                printf("usage: t <output dir>\n");

            continue;
        }
        if (*cmd == 'u')
        {
            char url[255];

            if (sscanf(cmd, " %*c %[^\n]", url) == 1)
            {
                server = url;
                pimp.setServer(server.c_str(), 80);
                printf("New server is [%s]\n", url);
            }
            else
                printf("usage: u <output dir>\n");

            continue;
        }
        if (*cmd == 'x')
        {
            clearTagsOpt = !clearTagsOpt;
            pimp.setDebug(debugOpt);
            printf("Clear tags: %s\n", clearTagsOpt ? "on" : "off");
            pimp.setClearTags(clearTagsOpt);
            continue;
        }
        if (strncmp(cmd, "fv", 2) == 0)
        {
            char mask[255];

            if (sscanf(cmd, " %*c %[^\n]", mask) == 1)
            {
                variousFileMask = mask;
                pimp.setVariousFileMask(variousFileMask);
                printf("New various file mask is [%s]\n", mask);
            }
            else
                printf("usage: fv <various file mask>\n");

            continue;
        }
        if (*cmd == 'f')
        {
            char mask[255];

            if (sscanf(cmd, " %*c %[^\n]", mask) == 1)
            {
                fileMask = mask;
                pimp.setFileMask(fileMask);
                printf("New file mask is [%s]\n", mask);
            }
            else
                printf("usage: f <file mask>\n");

            continue;
        }
        if (*cmd == 'h' || *cmd == '?')
        {
            printf("Help:\n"
                   "Command:  Arguments:        Function:\n"
                   "---------------------------------------------------------\n"
                   "   a      <filename>        Add file\n"
                   "   d      <dir>             Add files in dir\n"
                   "   l                        List files in tagger\n"
                   "   p      <file #>          Print details of file\n"
                   "   r      <file #>          Show the matches for a file lookup\n"
                   "   q                        Quit\n"
                   "   w      [file #s]         Write tags & rename files\n"
                   "   c      <file #> <choice> Choose a search result\n"
                   "   s                        Submit TRMs to server\n"
                   "   i      <file #>          Identify file again\n"
                   "   m      <file #>          File was misidentified\n"
                   "   n      <file #>          Provide new metadata for file\n"
                   "   v      <file #>          Toggle Various Artist flag for file\n\n"
                   "Options:\n"
                   "---------------------------------------------------------\n"
                   "   e                        Toggle file rename (%s)\n"
                   "   o                        Toggle file move (%s)\n"
                   "   t      <output dir>      Set output dir (%s)\n"
                   "   u      <server name>     Use MB server (%s)\n"
                   "   tc     <#>               Set TRM collision threshold (%d)\n"
                   "   ta     <#>               Set auto save threshold (%d)\n"
                   "   x                        Clear tags before writing new ones (%s)\n"
                   "   f      <mask>            Set the filemask (%s)\n"
                   "   fv     <va mask>         Set the various artist filemask (%s)\n"
                   "\n", 
                   renameOpt ? "on" : "off",
                   moveOpt ? "on" : "off",
                   outputDir.c_str(),
                   server.c_str(),
                   collThreshold,
                   autoSaveThreshold,
                   clearTagsOpt ? "on" : "off",
                   fileMask.c_str(),
                   variousFileMask.c_str()
                   );
            continue;
        }
        printf("Unknown command: %d\n", *cmd);
    }
}
