/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) Robert Kaye 2003
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: tunepimp.h,v 1.19 2003/06/23 09:27:19 robert Exp $

----------------------------------------------------------------------------*/
#ifndef __TUNEPIMP_H__
#define __TUNEPIMP_H__

#include <vector>
#include <map>
using namespace std;

#include "mutex.h"
#include "track.h"
#include "context.h"
#include "filecache.h"
#include "analyzer.h"
#include "submit.h"
#include "lookup.h"
#include "filelookup.h"
#include "write.h"

enum TPError
{
    tpOk,
    tpTooManyTRMs,
    tpNoUserInfo,
    tpLookupError,
    tpSubmitError,
    tpInvalidIndex,

    tpErrorLast
}; 

enum TPCallbackEnum
{
    tpFileAdded,
    tpFileChanged,
    tpFileRemoved,

    tpCallbackLast
};

class TPCallback
{
    public:

                  TPCallback(void) {};
        virtual  ~TPCallback(void) {};

        virtual   void notify(TPCallbackEnum, int fileId) = 0;
        virtual   void status(const string &status) = 0;
};

class TunePimp
{
    public:

                  TunePimp(const string &appName, const string &appVersion);
        virtual  ~TunePimp(void);

        // Setup functions -------------------------------------------------
        void      getVersion            (int &major, int &minor, int &rev);
        void      setUserInfo           (const string &userName, const string &password);
        bool      getUseUTF8            (void);
        void      setUseUTF8            (bool use);
                      
        void      setServer             (const string &server, short port);
        void      setProxy              (const string &proxyServer, short proxyPort);
        void      getServer             (string &server, short &port);
        void      getProxy              (string &proxyServer, short proxyPort);

        void      getSupportedExtensions(vector<string> &extList);
        void      setAutoFileLookup     (bool enable);
        void      setCallback           (TPCallback *callback);

        // When a function fails, call this function to get an error string.
        void      getError              (string &error);

        // Call with true if you want to print server query/response info
        void      setDebug              (bool debug);


        // File functions --------------------------------------------------
        
        // Returns a fileId number that can be used to reference the tracks later
        int       add                   (const string &fileName);
        int       addDir                (const string &dirPath);
        void      remove                (int fileId);
                
        int       getNumFiles           (void);
        int       getNumUnsubmitted     (void);
        void      getFileIds            (vector<int> &ids);
        Track    *getTrack              (int fileId);
        void      releaseTrack          (Track *track);

        // User feedback functions -----------------------------------------
        TPError   selectResult          (Track *track, int resultIndex);
        void      misidentified         (int fileId);
        void      identifyAgain         (int fileId);

        // Writing & submitting functions ----------------------------------
        bool      writeTags             (vector<int> *fileIds = NULL);
        TPError   submitTRMs            (void);

        // Config options ----------------------------------
        void      setRenameFiles          (bool rename);
        void      setMoveFiles            (bool move);
        void      setWriteID3v1           (bool writeID3v1);
        void      setWriteID3v2           (bool writeID3v2);
        void      setClearTags            (bool clearTags);
        void      setFileMask             (const string &fileMask);
        void      setVariousFileMask      (const string &variousFileMask);
        void      setDestDir              (const string &destDir);
        void      setTopSrcDir            (const string &topSrcDir);
        void      setTRMCollisionThreshold(int trmThreshold);
        void      setAutoSaveThreshold    (int autoSaveThreshold);

        // Stoopid windows socket setup functions --------------------------
#ifdef WIN32
        void      WSAInit               (void);
        void      WSAStop               (void);
#endif

        // This is public for easy access. Everyone should have access to it.
        Context           context;

    private:

        FileCache         cache;
        Analyzer         *analyzer;
        WriteThread      *write;
        LookupThread     *lookup;
        FileLookupThread *fileLookup;
        SubmitInfo       *submit;
        string            server, proxyServer;
        short             port, proxyPort;
        bool              useUTF8;
        string            err;
        vector<string>    extList;
        TPCallback       *callback;

    public:

        // These functions are for internal use only
        void     wake      (Track *track);
        void     setStatus (const string &status);
};

#endif
