/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) Robert Kaye 2003
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: analyzer.cpp,v 1.16 2003/06/27 23:08:25 robert Exp $

----------------------------------------------------------------------------*/
#ifdef WIN32
#if _MSC_VER == 1200
#pragma warning(disable:4786)
#endif
#endif

#include <stdio.h>
#include "../config.h"
#include "analyzer.h"
#include "tunepimp.h"
#include "submit.h"
#include "file_meta.h"
#include "mp3_trm.h"
#include "ov_trm.h"
#include "wav_trm.h"
#include "id3_meta.h"
#include "vorbis_meta.h"
#include "ape_trm.h"
#include "ape_meta.h"

extern "C"
{
   #include "fstrcmp.h"
}   

const int iDataFieldLen = 255;

#define DB printf("%s:%d\n", __FILE__, __LINE__);

//---------------------------------------------------------------------------

Analyzer::Analyzer(TunePimp   *tunePimpArg,
                   FileCache  *cacheArg,
                   SubmitInfo *submitInfoArg) : Thread()
{
   tunePimp = tunePimpArg;
   cache = cacheArg;
   submitInfo = submitInfoArg;
   // TODO: Allow thread priority setting
   // Priority = tpLowest;
   exitThread = false;
   sem = new Semaphore("analyzer");
}

//---------------------------------------------------------------------------

Analyzer::~Analyzer(void)
{
   exitThread = true;
   sem->signal();
   join();
   delete sem;
}

//---------------------------------------------------------------------------

void Analyzer::wake(void)
{
    sem->signal();
}

//---------------------------------------------------------------------------

void Analyzer::threadMain(void)
{
    string  fileName, status, trm;
    Track  *track;

    for(; !exitThread;)
    {
        track = cache->getNextItem(ePending);
        if (track == NULL)
        {
           sem->wait();
           continue;
        }

        track->lock();
        
        track->getFileName(fileName);
        status = "Analyzing " + fileName;
        tunePimp->setStatus(status);

        readMetadata(track);

        // Check to see if we pulled up a musicbrainz id
        Metadata data, id3;
        track->getServerMetadata(data);
        track->getLocalMetadata(id3);
        track->getTRM(trm);
        if (trm == string(ANALYZER_REDO_STRING))
        {
            data.clear();
            data.trackId = "";
            track->setServerMetadata(data);

            track->getLocalMetadata(id3);
            id3.trackId = "";
            track->setLocalMetadata(id3);
        }
        trm = "";

        if (!id3.trackId.empty())
        {
            // Yup, toss it into categorized
            status = fileName + " was previously matched.";
            tunePimp->setStatus(status);
            track->getLocalMetadata(data);
            track->setServerMetadata(data);
            track->setStatus(eRecognized);
        }
        else
        {
            unsigned long  duration = 0;
            char          *ptr;
            TRMGenerator  *gen;
            TRMResult      ret;


            ptr = strrchr(fileName.c_str(), '.');
#ifdef HAVE_LIBMAD
            if (ptr && strcasecmp(ptr, ".mp3") == 0)
                gen = new TRMGeneratorMP3();
            else
#endif
#ifdef HAVE_OGGVORBIS
            if (ptr && strcasecmp(ptr, ".ogg") == 0)
                gen = new TRMGeneratorOggVorbis();
            else
#endif
#ifdef HAVE_APE
            if (ptr && strcasecmp(ptr, ".ape") == 0)
                gen = new TRMGeneratorApe();
            else
#endif
            if (ptr && strcasecmp(ptr, ".wav") == 0)
                gen = new TRMGeneratorWav();
            else
                gen = NULL;


            if (gen)
            {
                gen->setProxy(proxyServer, proxyPort);
                track->unlock();
                try
                {
                    ret = gen->generate(fileName.c_str(), trm, duration);
                }
                catch(...)
                {
                    ret = eDecodeError;
                }
                track->lock();

                if (ret != eOk)
                {
                    string err;

                    tunePimp->setStatus("Failed to generate trm from " + fileName);
                    gen->getError(err);
                    if (err.empty())
                        setError(track, ret);
                    else
                        track->setError(err);
                    track->setStatus(eError);
                }

                delete gen;

                if (duration > 0)
                {
                    Metadata fileData;

                    track->getLocalMetadata(fileData);
                    fileData.duration = duration;
                    track->setLocalMetadata(fileData);
                }

                if (!trm.empty())
                {
                    if (trm.substr(14, 1) != string("4"))
                    {
                        string err;

                        err = string("Unable to retrieve TRM from fingerprint server.");
                        tunePimp->setStatus(err);
                        track->setStatus(eError);
                        track->setError(err);
                    }
                    else
                    {
                        track->setTRM(trm);
                        track->setStatus(eTRMLookup);
                    }
                }
            }
            else
            {
                string err;
                if (ptr)
                    err = string(ptr) + string(" is not a supported filetype.");
                else
                    err = string(fileName) + string(": cannot determine filetype.");
                tunePimp->setStatus(err);
                track->setStatus(eError);
                track->setError(err);
            }
        }
        track->unlock();
        tunePimp->wake(track);
        cache->release(track);
    }
}

//---------------------------------------------------------------------------

void Analyzer::setError(Track *track, TRMResult retVal)
{
    switch(retVal)
    {
        case eFileNotFound:
           track->setError("Audio file not found.");
           break;
        case eDecodeError:
           track->setError("Cannot decode audio file.");
           break;
        case eCannotConnect:
           track->setError("Cannot connect to the signature server.");
           break;
        default:
           track->setError("Unknown error. Sorry, this program sucks.");
           break;
    }
}

//---------------------------------------------------------------------------

// This whole scheme should get replaced with something more dynamic.
// But that takes a lot of time and effort, which I don't really have
// right now. But then again, do you ever?
void Analyzer::getSupportedExtensions(vector<string> &extList)
{
    extList.push_back(string(".wav"));
#ifdef HAVE_LIBMAD
    extList.push_back(string(".mp3"));
#endif
#ifdef HAVE_OGGVORBIS
    extList.push_back(string(".ogg"));
#endif
#ifdef HAVE_APE
    extList.push_back(string(".ape"));
#endif
}

//---------------------------------------------------------------------------
// note: The track is locked when this function is called.
bool  Analyzer::readMetadata(Track *track)
{
    string        fileName, trm, ext;
    Metadata      data;
    char         *ptr;
    FileMetadata *fileMeta = NULL;
    bool          ret, readRet;

    track->getFileName(fileName);
    ptr = strrchr(fileName.c_str(), '.');
#ifdef HAVE_LIBMAD
    if (ptr && strcasecmp(ptr, ".mp3") == 0)
        fileMeta = new ID3(true);
    else
#endif
#ifdef HAVE_OGGVORBIS
    if (ptr && strcasecmp(ptr, ".ogg") == 0)
        fileMeta = new Vorbis();
    else
#endif
#ifdef HAVE_APE
    if (ptr && strcasecmp(ptr, ".ape") == 0)
        fileMeta = new Ape();
    else
#endif
        return false;

    track->unlock();
    readRet = fileMeta->read(fileName, data);
    track->lock();
    if (!readRet)
       ret = false;
    else
    {
       track->getTRM(trm);
       if (trm != string(ANALYZER_REDO_STRING) && !data.fileTrm.empty())
          track->setTRM(data.fileTrm);

       track->setLocalMetadata(data);
       ret = true;
    }

    delete fileMeta;
    
    return ret;
}


