/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) Robert Kaye 2003
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: tunepimp.cpp,v 1.27 2003/06/27 23:08:26 robert Exp $

----------------------------------------------------------------------------*/
#ifdef WIN32
#if _MSC_VER == 1200
#pragma warning(disable:4786)
#endif
#endif

#include <stdio.h>
#include <map>
using namespace std;

#ifdef WIN32
#include <winsock.h>
#endif

#include "../config.h"
#include "tunepimp.h"
#include "dirsearch.h"

#define DB printf("%s:%d\n", __FILE__, __LINE__);


TunePimp::TunePimp(const string &appName, const string &appVersion)
{
    callback = NULL;
    useUTF8 = true;

    submit = new SubmitInfo(this, appName, appVersion);
    analyzer = new Analyzer(this, &cache, submit);
    analyzer->getSupportedExtensions(extList);
    lookup = new LookupThread(this, &cache, submit);
    fileLookup = new FileLookupThread(this, &cache);
    write = new WriteThread(this, &cache);

    analyzer->start();
    lookup->start();
    fileLookup->start();
    write->start();
}

TunePimp::~TunePimp(void)
{
    Analyzer         *aTemp;
    LookupThread     *lTemp;
    FileLookupThread *fTemp;
    WriteThread      *rTemp;

    aTemp = analyzer;
    analyzer = NULL;
    delete aTemp;

    lTemp = lookup;
    lookup = NULL;
    delete lTemp;

    fTemp = fileLookup;
    fileLookup = NULL;
    delete fTemp;

    rTemp = write;
    write = NULL;
    delete rTemp;

    delete submit;
}

// Get the version of the library
void TunePimp::getVersion(int &major, int &minor, int &rev)
{
    sscanf(VERSION, "%d.%d.%d", &major, &minor, &rev);
}

bool TunePimp::getUseUTF8(void)      
{ 
    return useUTF8; 
}

void TunePimp::setUseUTF8(bool use) 
{ 
    useUTF8 = use; 
}
         
void TunePimp::setServer(const string &server, short port)
{
    this->server = server;
    this->port = port;
}

void TunePimp::setProxy(const string &proxyServer, short proxyPort)
{
    this->proxyServer = proxyServer;
    this->proxyPort = proxyPort;
}

void TunePimp::getServer(string &server, short &port)
{
    server = this->server;
    port = this->port;
}

void TunePimp::getProxy(string &proxyServer, short proxyPort)
{
    proxyServer = this->proxyServer;
    proxyPort = this->proxyPort;
}

void TunePimp::getSupportedExtensions(vector<string> &extList)
{ 
    extList = this->extList; 
}

void TunePimp::setWriteID3v1(bool writeID3v1)
{ 
    context.setWriteID3v1(writeID3v1);
}                           

void TunePimp::setClearTags(bool clearTags)
{ 
    context.setClearTags(clearTags);
}                           

void TunePimp::setDestDir(const string &destDir)
{ 
    context.setDestDir(destDir);
}                           

void TunePimp::setFileMask(const string &fileMask)
{ 
    context.setFileMask(fileMask);
}                           

void TunePimp::setVariousFileMask(const string &variousFileMask)
{ 
    context.setVariousFileMask(variousFileMask);
}                           

void TunePimp::setTopSrcDir(const string &srcDir)
{ 
    context.setTopSrcDir(srcDir);
}                           

void TunePimp::setMoveFiles(bool moveFiles)
{ 
    context.setMoveFiles(moveFiles);
}                           

void TunePimp::setRenameFiles(bool renameFiles)
{ 
    context.setRenameFiles(renameFiles);
}                           

void TunePimp::setTRMCollisionThreshold(int trmThreshold)
{
    context.setTRMCollisionThreshold(trmThreshold);
}

void TunePimp::setAutoSaveThreshold(int autoSaveThreshold)
{
    context.setAutoSaveThreshold(autoSaveThreshold);
}

void TunePimp::setAutoFileLookup(bool enable)
{
    if (lookup)
        lookup->setAutoFileLookup(enable);
}

void TunePimp::setCallback(TPCallback *callback) 
{ 
    this->callback = callback; 
}

void TunePimp::getError(string &error) 
{ 
    error = err; 
}

// Set the debug option -- if enabled debug output will be printed to stdout
void TunePimp::setDebug(bool debug)
{
    context.setDebug(debug);
}

void TunePimp::setUserInfo(const string &userName, const string &password)
{
    submit->setUserInfo(userName, password);
}

// Stoopid windows helper functions to init the winsock layer.
#ifdef WIN32
void TunePimp::WSAInit(void)
{
    WSADATA sGawdIHateMicrosoft;
    WSAStartup(0x0002,  &sGawdIHateMicrosoft);
}   

void TunePimp::WSAStop(void)
{
    WSACleanup();
}
#endif

int TunePimp::add(const string &fileName)
{
    int fileId;
    
    fileId = cache.add(fileName);
    analyzer->wake();
    if (callback)
        callback->notify(tpFileAdded, fileId);

    return fileId;
}

int TunePimp::addDir(const string &dirPath)
{
    DirSearch search(extList);
    int       count, fileId;

    count = search.recurseDir(dirPath.c_str());
    if (count > 0)
    {
        vector<string>           fileList;
        vector<string>::iterator i;
        search.getFiles(fileList);
        for(i = fileList.begin(); i != fileList.end(); i++)
        {
            fileId = cache.add(*i);
            if (callback)
                callback->notify(tpFileAdded, fileId);
        }
        analyzer->wake();
    }
    return count;
}

int TunePimp::getNumUnsubmitted(void)
{
    return submit->getNumItems();
}

void TunePimp::remove(int fileId)
{
    cache.remove(fileId);
    if (callback)
        callback->notify(tpFileRemoved, fileId);
}

void TunePimp::getFileIds(vector<int> &ids) 
{ 
    cache.getFileIds(ids); 
}

int TunePimp::getNumFiles(void)
{
    return cache.getNumItems();
}

Track *TunePimp::getTrack(int fileId)
{
    return cache.getTrack(fileId);
}

void TunePimp::releaseTrack(Track *track)
{
    cache.release(track);
}

// Since this function blocks, the track is assumed to be locked
// and should not be unlocked.
void TunePimp::wake(Track *track)
{
    if (callback)
    {
        int fileId;
        
        fileId = cache.getFileIdFromTrack(track);
        if (fileId >= 0)
            callback->notify(tpFileChanged, fileId);
    }

    if (analyzer)
        analyzer->wake(); 

    if (lookup)
        lookup->wake(); 

    if (fileLookup)
        fileLookup->wake(); 

    if (write)
        write->wake(); 
}

void TunePimp::setStatus(const string &status)
{
    if (callback)
        callback->status(status);
}

TPError TunePimp::submitTRMs(void)
{
    LookupStatus ret;

    ret = submit->submit();
    if (ret != eSubmitOk)
    {
        submit->getError(err);
        return tpSubmitError;
    }
    submit->clear();

    return tpOk;
}

bool TunePimp::writeTags(vector<int> *fileIds)
{
    vector<Track *>            tracks;
    vector<Track *>::iterator  i;
    vector<int>::iterator      j;
    Track                     *track;

    if (fileIds)
    {
        for(j = fileIds->begin(); j != fileIds->end(); j++)
        {
            track = cache.getTrack(*j);
            if (!track)
            {
                err = "Invalid track in write tags list.";
                return false;
            }
            if (track->getStatus() != eRecognized)
            {
                err = "All tracks must be recognized before writing tags.";
                return false;
            }
            tracks.push_back(track);
        }
    }
    else
        cache.getTracksFromStatus(eRecognized, tracks);

    for(i = tracks.begin(); i != tracks.end(); i++)
    {
        (*i)->lock();
        (*i)->setStatus(eVerified);
        (*i)->unlock();
        cache.release(*i);
    }
    write->wake();

    return true;
}


TPError TunePimp::selectResult(Track *track, int resultIndex)
{
    vector<TPResult *>  results;
    LookupFile::Result  type;
    TPResult           *result;
    Metadata            data;

    track->lock();

    track->getResults(type, results);
    if (resultIndex < 0 || resultIndex >= (int)results.size())
    {
        track->unlock();
        return tpInvalidIndex;
    }

    result = results[resultIndex];

    track->getServerMetadata(data);
    if (type == LookupFile::eArtistList)
    {
        TPArtistResult *res = (TPArtistResult *)result;

        data.artistId = res->getId();
        data.artist = res->getName();
        data.sortName = res->getSortName();

        track->setServerMetadata(data);
        track->setStatus(eFileLookup);
    }
    else
    if (type == LookupFile::eAlbumList)
    {
        TPAlbumResult *res = (TPAlbumResult *)result;
        TPArtistResult artist;
        
        artist = res->getArtist();

        data.artistId = artist.getId();
        data.artist = artist.getName();
        data.sortName = artist.getSortName();

        data.albumId = res->getId();
        data.album = res->getName();
        data.variousArtist = res->getVariousArtists();

        track->setServerMetadata(data);

        // If we have track information, try and resolve that as well.
        // If not, leave things here -- there is nothing else to do.
        track->getLocalMetadata(data);
        if (!data.track.empty())
            track->setStatus(eFileLookup);
        else
            track->setStatus(eUnrecognized);
    }
    else
    if (type == LookupFile::eTrackList)
    {
        string trm;
        TPArtistResult artist;
        TPAlbumResult album;

        TPAlbumTrackResult *res = (TPAlbumTrackResult *)result;
        
        artist = res->getArtist();
        album = res->getAlbum();

        data.artistId = artist.getId();
        data.artist = artist.getName();
        data.sortName = artist.getSortName();

        data.albumId = album.getId();
        data.album = album.getName();
        data.variousArtist = album.getVariousArtists();

        data.trackId = res->getId();
        data.track = res->getName();
        data.duration = res->getDuration();
        data.trackNum = res->getTrackNum();

        track->getTRM(trm);
        submit->add(data.trackId, trm);

        track->setServerMetadata(data);
        track->setStatus(eRecognized);
    }

    wake(track); 
    track->unlock();

    return tpOk;
}

void TunePimp::misidentified(int fileId)
{
    Track *track;

    track = cache.getTrack(fileId);
    if (track)
    {
        string   trm;
        Metadata data;

        track->lock();
        track->getTRM(trm);
        track->getServerMetadata(data);
        if (!data.trackId.empty() && !trm.empty())
            submit->remove(data.trackId);

        if (trm.empty())
        {
            track->setTRM(ANALYZER_REDO_STRING);
            track->setStatus(ePending);
        }
        else
            track->setStatus(eFileLookup);

        data.clear();
        track->setServerMetadata(data);
        track->setError("");
        track->unlock();
        wake(track);
        cache.release(track);
    }
}

void TunePimp::identifyAgain(int fileId)
{
    Track *track;

    track = cache.getTrack(fileId);
    if (track)
    {
        string   trm;
        Metadata data;

        track->lock();
        track->getTRM(trm);
        track->getServerMetadata(data);
        if (!data.trackId.empty() && !trm.empty())
            submit->remove(data.trackId);

        data.clear();
        track->setTRM(ANALYZER_REDO_STRING);
        track->setServerMetadata(data);
        track->setError("");
        track->setStatus(ePending);
        track->unlock();
        wake(track);
        cache.release(track);
    }
}
