/*----------------------------------------------------------------------------

   libtunepimp -- The MusicBrainz tagging library.  
                  Let a thousand taggers bloom!
   
   Copyright (C) Robert Kaye 2003
   
   This file is part of libtunepimp.

   libtunepimp is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   libtunepimp is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with libtunepimp; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   $Id: write.cpp,v 1.15 2003/06/27 23:08:26 robert Exp $

----------------------------------------------------------------------------*/
#ifdef WIN32
#  if _MSC_VER == 1200
#		pragma warning(disable:4786)
#	endif
#  define  strcasecmp stricmp 
#else
#  include <unistd.h>
#  include <sys/stat.h>
#  include <sys/types.h>
#  include <fcntl.h>
#  if defined(__APPLE__) || defined(__NetBSD__)
#    include <sys/param.h>
#    include <sys/mount.h>
#  else
#    include <sys/vfs.h>
#  endif
#endif


#include <stdio.h>
#include <ctype.h>
#include <errno.h>
using namespace std;

#include "../config.h"
#include "write.h"
#include "tunepimp.h"
#include "file_meta.h"
#include "id3_meta.h"
#include "vorbis_meta.h"
#include "ape_meta.h"

#ifdef WIN32
const char *dirSep = "\\";
const char dirSepChar = '\\';
const char *disallowedFileNameChars = "\"*/:<>?|";
#else
const char *dirSep = "/";
const char dirSepChar = '/';
#endif

#define DB printf("%s:%d\n", __FILE__, __LINE__);

//---------------------------------------------------------------------------

WriteThread::WriteThread(TunePimp  *tunePimpArg,
                         FileCache *cacheArg) : Thread()
{
    tunePimp = tunePimpArg;
    cache = cacheArg;

    // TODO: Allow thread priority setting
    // Priority = tpLowest;
    exitThread = false;
    sem = new Semaphore("write");
}

//---------------------------------------------------------------------------

WriteThread::~WriteThread(void)
{
    exitThread = true;
    sem->signal();
    join();
    delete sem;
}

//---------------------------------------------------------------------------

void WriteThread::wake(void)
{
    sem->signal();
}

//---------------------------------------------------------------------------

void WriteThread::threadMain(void)
{
    Metadata  server;
    string    fileName, status, trm, trackId;
    Track    *track;

    for(; !exitThread;)
    {
        track = cache->getNextItem(eVerified);
        if (track == NULL)
        {
           sem->wait();
           continue;
        }

        track->lock();
        track->getServerMetadata(server);
        track->getTRM(server.fileTrm);

        if (track->hasChanged() || renameCheck(track, server))
        {
            string fileName;

            track->getFileName(fileName);
            track->unlock();
            if (writeTrack(track, server))
            {
                track->lock();
                track->setStatus(eDeleted);
                track->setError("Track saved.");
            }
            else
            {
                track->lock();
                track->setStatus(eError);
            }
        }
        else
        {
            string fileName;

            track->getFileName(fileName);
            track->setStatus(eDeleted);
        }

        track->unlock();

        tunePimp->wake(track);
        cache->release(track);
    }
}

//---------------------------------------------------------------------------

bool WriteThread::writeTrack(Track *track, const Metadata &server)
{
    string           ext, fileName;
    FileMetadata    *fileMeta = NULL;
    unsigned long    fileSize;

    track->lock();
    track->getFileName(fileName);
    ext = extractFileExt(fileName);

    track->unlock();
    fileSize = fileOpenTest(fileName);
    track->lock();

    if (fileSize == 0)
    {
        track->setError("Cannot open file for writing.");
        track->unlock();
        return false;
    }
    
    track->unlock();
    if (!diskSpaceTest(fileName, fileSize))
    {
        track->lock();
        track->setError("Not enough available diskspace for writing a new file.");
        track->unlock();
        return false;
    }

#ifdef HAVE_LIBMAD
    if (strcasecmp(ext.c_str(), ".mp3") == 0)
        fileMeta = new ID3(tunePimp->context.getWriteID3v1());
    else
#endif
#ifdef HAVE_OGGVORBIS
    if (strcasecmp(ext.c_str(), ".ogg") == 0)
        fileMeta = new Vorbis();
    else
#endif
#ifdef HAVE_APE
    if (strcasecmp(ext.c_str(), ".ape") == 0)
        fileMeta = new Ape();
    else
#endif
        fileMeta = NULL;

    if (fileMeta)
    {
        bool   ret;
        string err;

        try
        {
            ret = fileMeta->write(fileName, server, tunePimp->context.getClearTags());
        }
        catch(...)
        {
            ret = false;
        }
        fileMeta->getError(err);
        delete fileMeta;

        if (!ret)
        {
            track->lock();
            track->setError(string("Could not write metadata to track: ") + err);
            track->unlock();
            return false;
        }
    }

    if (tunePimp->context.getRenameFiles() && tunePimp->context.getMoveFiles())
    {
        string newName, err;
        int    ret;

        for(int j = 0;; j++)
        {
            newName = fileName;
            makeNewFileName(server, newName, j);
            if (createPath(newName))
            {
                if (newName == fileName)
                   break;

                ret = rename(fileName.c_str(), newName.c_str());
                if (ret != 0 && errno == EEXIST)
                    continue;

                if (ret != 0)
                {
                    track->lock();
                    track->setError("Could not rename file.");
                    track->unlock();
                    return false;
                }

                if (tunePimp->context.getMoveFiles())
                    cleanPath(fileName);
            }
            else
            {
                string path = extractFilePath(newName);

                err = string("Could not create destination directory: ") + path;
                track->lock();
                track->setError(err);
                track->unlock();
                return false;
            }

            break;
        }
        track->lock();
        track->setFileName(newName);
        track->unlock();
    }

    return true;
}

//---------------------------------------------------------------------------

bool WriteThread::renameCheck(Track *track, const Metadata &data)
{
    string origFileName, newFileName;

    if (!tunePimp->context.getRenameFiles())
        return false;

    track->getFileName(origFileName);
    makeNewFileName(data, newFileName, 0);

    return strcmp(newFileName.c_str(), origFileName.c_str()) != 0; 
}

//---------------------------------------------------------------------------

#ifdef WIN32
unsigned long WriteThread::fileOpenTest(const string &fileName)
{
    HANDLE        openTest;
    unsigned long fileSize;
    
    openTest = CreateFile(fileName.c_str(), GENERIC_READ | GENERIC_WRITE,
                          0, NULL, OPEN_EXISTING, 0, NULL);
    if (openTest == INVALID_HANDLE_VALUE)
        return 0;

    fileSize = SetFilePointer(openTest, 0, NULL, FILE_END);
    CloseHandle(openTest);

    return fileSize;
}

#else
//---------------------------------------------------------------------------

unsigned long WriteThread::fileOpenTest(const string &fileName)
{
    int           openTest;
    unsigned long fileSize;
   
    // This test is probably not NFS 
    openTest = open(fileName.c_str(), O_EXCL | O_RDWR);
    if (openTest < 0)
        return 0;

    fileSize = lseek(openTest, 0, SEEK_END);
    close(openTest);

    return fileSize;
}
#endif

//---------------------------------------------------------------------------

#ifdef WIN32
bool WriteThread::diskSpaceTest(const string &fileName, unsigned long fileSize)
{
    ULARGE_INTEGER temp; 
    __int64        diskSpace;

    string path = extractFilePath(fileName);
    path += "\\";

    GetDiskFreeSpaceEx(path.c_str(), &temp, NULL, NULL);
    diskSpace = *(__int64 *)&temp;

    // Increase the size a bit in case the file grows and what not.
    fileSize += fileSize / 10;
    return diskSpace > (__int64)fileSize;
}

#else
//---------------------------------------------------------------------------

bool WriteThread::diskSpaceTest(const string &fileName, unsigned long fileSize)
{
    struct statfs stat;

    if (statfs(fileName.c_str(), &stat) == 0)
    {
        fileSize += fileSize / 10;
        fileSize /= stat.f_bsize;

        return fileSize < (unsigned long)stat.f_bavail;
    }
    else
        return false;
}
#endif

//---------------------------------------------------------------------------

string WriteThread::extractFilePath(const string &file)
{
    string::size_type pos;
    
    pos = file.rfind(dirSep, file.size() - 1);
    if (pos == string::npos)
        return file;

    return file.substr(0, pos);
}

//---------------------------------------------------------------------------

string WriteThread::extractFileName(const string &file)
{
    string::size_type pos;
    
    pos = file.rfind(dirSep, file.size() - 1);
    if (pos == string::npos)
        return file;

    return file.substr(pos + 1, file.size());
}

//---------------------------------------------------------------------------

string WriteThread::extractFileExt(const string &file)
{
    string::size_type pos;
    
    pos = file.rfind(".", file.size() - 1);
    if (pos == string::npos)
        return file;

    return file.substr(pos, file.size());
}

//---------------------------------------------------------------------------

string WriteThread::extractVolume(const string &file)
{
#ifdef WIN32
    string::size_type pos;

    if (file.size() > 2 && file[0] == '\\' && file[1] == '\\')
    {
        pos = file.find("\\", 2);
        if (pos == string::npos)
            return "";

        return file.substr(0, pos);
    }

    if (file.size() > 2 && isalpha(file[0]) && file[1] == ':')
    {
        pos = file.find("\\", 3);
        if (pos == string::npos)
            return "";

        return file.substr(0, pos);
    }
#endif
   
    return "";
}

//---------------------------------------------------------------------------

const string WriteThread::sanitize(const string &str)
{
    string data;

    data = str;
    for(unsigned i = 0; i < str.size(); i++)
       if (str[i] == dirSepChar)
           data.erase(i, 1);

    return data;
}

//---------------------------------------------------------------------------

void WriteThread::makeNewFileName(const Metadata &dataArg,
                                  string         &fileName,
                                  int             index)
{
    Metadata          data = dataArg;
    string            name, origPath, ext;
    string::size_type pos;

    origPath = extractFilePath(fileName);
    ext = extractFileExt(fileName);
    if (data.variousArtist)
        name = tunePimp->context.getVariousFileMask();
    else
        name = tunePimp->context.getFileMask();

    if (!tunePimp->context.getRenameFiles() && tunePimp->context.getMoveFiles())
    {
        fileName = tunePimp->context.getDestDir() + string(dirSep) + 
                   extractFileName(fileName);
        return;
    }

    if (data.sortName.empty())
        data.sortName = data.artist;

    for(;;)
    {
        pos = name.find("%artist");
        if (pos != string::npos)
        {
             name.erase(pos, 7);
             name.insert(pos, data.artist);
             continue;
        }

        pos = name.find("%sortname");
        if (pos != string::npos)
        {
             name.erase(pos, 9);
             name.insert(pos, data.sortName);
             continue;
        }

        pos = name.find("%abc2");
        if (pos != string::npos)
        {
             name.erase(pos, 5);
             name.insert(pos, data.sortName.substr(0, 2));
             continue;
        }

        pos = name.find("%abc3");
        if (pos != string::npos)
        {
             name.erase(pos, 5);
             name.insert(pos, data.sortName.substr(0, 3));
             continue;
        }

        pos = name.find("%abc");
        if (pos != string::npos)
        {
             name.erase(pos, 4);
             name.insert(pos, data.sortName.substr(0, 1));
             continue;
        }

        pos = name.find("%album");
        if (pos != string::npos)
        {
             name.erase(pos, 6);
             name.insert(pos, data.album);
             continue;
        }

        pos = name.find("%0num");
        if (pos != string::npos)
        {
             char num[10];

             sprintf(num, "%02d", data.trackNum);
             name.erase(pos, 5);
             name.insert(pos, num);
             continue;
        }

        pos = name.find("%num");
        if (pos != string::npos)
        {
             char num[10];

             sprintf(num, "%d", data.trackNum);
             name.erase(pos, 4);
             name.insert(pos, num);
             continue;
        }

        pos = name.find("%track");
        if (pos != string::npos)
        {
             name.erase(pos, 6);
             name.insert(pos, data.track);
             continue;
        }
        break;
    }

#ifdef WIN32
    // Now remove any characters that the filesystem might bitch about
    for(unsigned i = 0; i < name.size(); i++)
    {
        if (strchr(disallowedFileNameChars, name[i]))
        {
            name.erase(i, 1);
            i--;
        }
    }
#endif

    if (index > 0)
    {
        char temp[10];

        sprintf(temp, " (%d)", index);
        fileName = name + string(temp) + string(ext);
    }
    else
        fileName = name + string(ext);

    if (tunePimp->context.getMoveFiles())
        fileName = tunePimp->context.getDestDir() + string(dirSep) + fileName;
    else
        fileName = origPath + fileName;
}


bool WriteThread::createPath(const string &pathArg)
{
    string            path = string(extractFilePath(pathArg).c_str());
    string            volume = string(extractVolume(pathArg).c_str());
    string            partial;
    string::size_type pos;


    if (volume.size() > 0)
        path.erase(0, volume.size());

    if (path[path.size() - 1] != dirSepChar)
        path += dirSep;

    for(pos = 1;;)
    {
        pos = path.find(dirSep, pos);
        if (pos == string::npos)
            break;

        partial = volume + path.substr(0, pos);
        if (access(partial.c_str(), 0))
        {
#ifdef WIN32
            if (mkdir(partial.c_str()) < 0)
#else
            if (mkdir(partial.c_str(), 0755) < 0)
#endif
                return false;
        }

        pos++;
    }

    return true;
}

void WriteThread::cleanPath(const string &pathArg)
{
    string      path = string(extractFilePath(pathArg).c_str());
    string      volume = string(extractVolume(pathArg).c_str());
    string      srcDir, complete;
    unsigned    pos;
    int         ret;

    srcDir = tunePimp->context.getTopSrcDir();
    if (volume.size() > 0)
        path.erase(0, volume.size());

    if (path[path.size() - 1] == dirSepChar)
        path.erase(path.size() - 1);

    if (srcDir[srcDir.size() - 1] == dirSepChar)
        srcDir.erase(srcDir.size() - 1);

    for(;;)
    {
        complete = volume + path;
        if (strcasecmp(srcDir.c_str(), complete.c_str()) == 0)
            break;
           
        ret = rmdir(complete.c_str());
        if (ret)
            break;

        pos = path.rfind(dirSep);
        if (pos == string::npos)
            break;

        path.erase(pos);
    }
}
